import rs, { IRequest, App, IAppOptions, Response, RouteAuth, ErrorField, HttpError, Param, Field, RSError, Type, Application } from './src';

const jr: Response.JSON = new Response.JSON();
class ImageResponse extends Response.Base {

	get() {
		return null;
	}

	getContentType() {
		return 'image/png';
	}

}

interface IEndpointProps {
	permissions: string;
}

interface R<A = any, Q = any, B = any> extends IRequest<A, Q, B, IEndpointProps> {
	user: {
		id: number;
	};
}

const createApp = (options?: IAppOptions): App => {
	return new Application(options);
};

const e = new RSError.InvalidParameterType('test', Type.string);
const notFound = new RSError.NotFound();

const app = rs({
	apiKey: {
		enabled: true,
		validator: (apiKey: string) => new Promise((resolve) => resolve(apiKey === 'API_KEY')),
	},
	log: {
		enabled: true,
		level: 'verbose',
		stack: true,
	},
	timeout: 30000,
	auth: {
		async validator(key, req): Promise<boolean> {
			return req.headers[key] === 'accessToken';
		},
	},
	charset: 'utf-8',
	logger(data, req) {
		console.log(data);
	},
});

app.get<R<{ a: string }, { b: string }, { c: string }>, IEndpointProps>(0, '/test', {
	excludedApiKeys: () => new Promise(resolve => resolve(['test'])),
	response: new Response.JSON(),
	auth: RouteAuth.REQUIRED,
	props: {
		permissions: 'test',
	},
}, (req, res, next) => {
	const { params, query, body, user } = req;
	req.getEndpoint().props.permissions;
	const b = req.getBenchmark();
	params.a;
	query.b;
	body.c;
	req.headers;
	res.send204();
	b.mark('endpoint execution');
	b.toJSON();
	next();
});

app.get(1, '/test', {
	excludedApiKeys: () => new Promise(resolve => resolve(['test'])),
	timeout: 15000,
}, async ({ params, body, query }) => null);

app.get(0, '/image', {
	response: new ImageResponse(),
}, (req, res, next) => next());

rs({
	apiKey: {
		enabled: true,
		validator(apiKey, next) {

		},
	},
	auth: {
		key: 'authorization',
		validator(key, req, res, next) {
			// req.user;
			res.send204();
		},
	},
});

app.start((err, { stats }) => {

});

const docs = app.getDocs();

docs['test'].response['test'].type;
docs['test'].response.shape['test'].type;

createApp({
	port: 666,
}).start();

app.stop();